      DOUBLE PRECISION FUNCTION DAQUAD(RULE,N,F,A,B,TOL,
     *					MAXLEV,sing,ERREST)

      DOUBLE PRECISION QRIGHT(1), XSAVE(1), F, DQUAD,
     *                 A, B, TOL, ERREST, FAC, SUM, FLAG,
     *                 XL, XR, TOLPU, XM, TEMP, sing
      EXTERNAL RULE,F

	double precision l_qright, l_qleft, l_qnow, l_qprev, l_qdiff,
     +			 r_qright, r_qleft, r_qnow, r_qprev, r_qdiff,
     +			 totalerr, err, totalval, val


c	*******************************
c	l_qright, l_qleft: values of integral, obtained by subdividing the
c			   left subinterval
c	l_qnow:	 value for left subinterval, obtained by adding l_qright &
c		 l_qleft
c	l_qprev: value for left subinterval, obtained by performing
c		 calulations on the interval as a whole
c	l_qdiff: difference between l_qnow and l_qprev, used to estimate
c		the error
c
c
c	r_qright, r_qleft, r_qnow, r_qprev, and r_qdiff and similarly
c	defined, but for the right subinterval
c
c
c	totalerr: the accumulated error for the entire interval to be
c		  integrated
c	err: 	  used to store the error of the subinterval being
c		  re-evaluated
c	totalval: the accumulated value of the integral for the entire
c		  interval
c	val:	  used to store the value of the subinterval being
c		  re-evaluated
c	*******************************


	integer heap(0:1000), pos

c	*******************************
c	heap: heap(0)->contains the number of elements in the heap
c	      heap(1:1000)->contain pointer to data stored in h_ arrays
c			  ->entries are ordered to form a max heap based
c			    on the h_error array
c	pos: used to remember the position in the heap of a removed
c	     element
c	*******************************


	double precision  h_left(1000), h_right(1000),
     +			  h_value(1000), h_error(1000), 
     +			  h_lvalue(1000), h_rvalue(1000)


c	*******************************
c	all h_ arrays belong to the max heap and are only accessed by
c	using the indices stored in the heap array
c
c	h_left(i), h_right(i): left and right limits for the i'th
c			       subinterval
c	h_value(i), h_error(i): current value and error for the i'th
c				subinterval
c	h_lvalue(i), h_rvalue(i): values for the left and right
c				  subintervals of the i'th subinterval
c	*******************************


C
C      ***************
C      *
C      * STRAIGHT INTEGRATION IF MAXLEV = 0.
C      *
C      ***************
C

         IF (MAXLEV .NE. 0) GO TO 4
              DAQUAD = DQUAD(RULE,N,F,A,B)
      RETURN
C
C      ***************
C      *
C      * OBTAIN  THE  APPROPRIATE SCALING FACTOR  FAC  (FOR THE
C      * CONVERGENCE CRITERION AND THE ERROR ESTIMATE) FROM THE
C      * SUBROUTINE DEFINED BY RULE.
C      *
C      ***************
C

4     	 NF = -N
         CALL RULE(NF,A,B,XSAVE,QRIGHT)
         FAC = XSAVE(1)

C
C      ***************
C      *
C      * INITIALIZATION AND CALCULATIONS FOR THE ZERO-TH LEVEL.
C      *
C      ***************
C

         SUM = 0.0D0
         ERREST = 0.0D0
	 err=0.0d0
	 totalerr = 0.0d0
         FLAG = 0.0D0
         XL = A
         XR = B
         tolpu = tol*fac
         l_qprev= DQUAD(RULE,N,F,A,B)
	 heap(0)=0

c	*******************************
c	perform calculations to determine an initial estimate for the
c	integral
c	*******************************


	xm = 0.5d0*(xl + xr)
	call calc_vals(l_qprev, xl, xr, rule, n, f,
     +                 l_qleft, l_qright, l_qnow, l_qdiff)
	 totalerr = l_qdiff
	 totalval=l_qnow
	 val=l_qnow
C
C      ***************
C      *
C      * TEST FOR INTERVAL CONVERGENCE AND CHECK IF THE MAXIMUM LEVEL
C      * IS REACHED.
C      *
C      ***************
C

	if (dabs(totalerr) .le. tolpu) then
	  heap(0)=1
	  go to 40
	endif

        l_qprev = l_qleft
	r_qprev = l_qright
	pos = 1
	err = l_qdiff

	heap(0)=1

c	*******************************
c	Successively bisect the subinterval with the largest error
c	estimate until the heap is full or the estimated error for the 
c	entire interval is within the specified tolerance
c	*******************************


10	call calc_vals (l_qprev, xl, xm, rule, n, f,
     +		  l_qleft, l_qright, l_qnow, l_qdiff)
	call calc_vals (r_qprev, xm, xr, rule, n, f,
     +		  r_qleft, r_qright, r_qnow, r_qdiff)

	  if (heap(0) .ge. 999) go to 20
	    if (dabs(l_qdiff) .gt. dabs(r_qdiff)) then
	      call insert_top(xl, xm, l_qnow, l_qdiff, l_qleft, l_qright,
     +			 heap, h_left, h_right, h_value, h_error, 
     +			 h_lvalue, h_rvalue, pos)
	      call insert(xm, xr, r_qnow, r_qdiff, r_qleft, r_qright,
     +			 heap, h_left, h_right, h_value, h_error, 
     +			 h_lvalue, h_rvalue)

	    else
	      call insert_top(xm, xr, r_qnow, r_qdiff, r_qleft, r_qright,
     +			 heap, h_left, h_right, h_value, h_error, 
     +			 h_lvalue, h_rvalue, pos)
	      call insert(xl, xm, l_qnow, l_qdiff, l_qleft, l_qright,
     +			 heap, h_left, h_right, h_value, h_error, 
     +			 h_lvalue, h_rvalue)
	    endif


	    totalerr=totalerr - err + l_qdiff +r_qdiff
	    totalval=totalval - val + l_qnow +r_qnow
            if (dabs(totalerr) .le. tolpu) go to 40

	call remove_top(xl, xr, val, err, l_qprev, r_qprev, heap,
     +		   h_left, h_right, h_value, h_error, h_lvalue, h_rvalue)


	xm=0.5d0*(xl+xr)
	go to 10

c	*******************************
c	heap limit was reached, record failure and return estimate
c	*******************************

20	flag = heap(0)

C
C      ***************
C      *
C      * ADD THE CORRECTION TERM FROM THE FIRST STEP OF RICHARDSON
C      * EXTRAPOLATION.  MAKE SURE ERREST IS NOT LESS THAN ROUNDOFF
C      * LEVEL.
C      *
C      ***************
C


40       DAQUAD = totalval + totalerr/FAC
         ERREST = dabs(totalerr/FAC)
         TOL = FLAG
         MAXLEV = heap(0)
         IF (ERREST .EQ. 0.0D0) RETURN
50       TEMP = DABS(DAQUAD) + ERREST
         IF (TEMP .NE. DABS(DAQUAD)) RETURN
         ERREST = 2.0D0*ERREST
         GO TO 50

      END
